import json
import logging
from collections import defaultdict
from dataclasses import dataclass, field, replace
from typing import Any, Callable, Dict, Iterable, List, Union

from beet import Context, DataPack, LootTable, NamespaceFile, NamespaceProxy
from beet.contrib.format_json import get_formatter
from beet.contrib.vanilla import Vanilla
from beet.core.file import JsonFile, JsonFileBase
from jsonpath_ng import parse

from .models import (
    SmithedAppendRule,
    SmithedInsertRule,
    SmithedJsonFile,
    SmithedMergeRule,
    SmithedModel,
    SmithedPrependRule,
    SmithedRemoveRule,
    SmithedRule,
    SmithedValueSource,
)

logger = logging.getLogger(__name__)


class WeldError(Exception):
    ...


class DependencyError(WeldError):
    ...


class WeldMergeError(WeldError):
    ...


@dataclass
class ConflictsHandler:
    ctx: Context

    formatter: Callable[..., str] = get_formatter()
    cache: defaultdict[type[JsonFile], set[str]] = field(
        default_factory=lambda: defaultdict(set)
    )
    vanilla: set[str] = field(default_factory=set)

    def __call__(
        self, pack: DataPack, path: str, current: JsonFile, conflict: JsonFile, /
    ):
        if (
            not path.startswith("minecraft:")
            or not path == "minecraft:blocks/yellow_shulker_box"
        ):
            return False
        smithed_current = SmithedJsonFile.parse_obj(current.data)
        smithed_conflict = SmithedJsonFile.parse_obj(conflict.data)

        if current_entries := smithed_current.smithed.entries():
            current_entries[-1].id = pack.mcmeta.data.get("id", "")

        if conflict_entries := smithed_conflict.smithed.entries():
            current_entries.extend(conflict_entries)

        json_file_type = type(current)
        self.cache[json_file_type].add(path)

        raw = json.loads(smithed_current.json(by_alias=True))
        if path not in self.vanilla:
            vanilla = self.ctx.inject(Vanilla)
            raw |= vanilla.data[json_file_type][path].data
            self.vanilla.add(path)

        pack[json_file_type][path] = json_file_type(raw)

        return True

    def handle_files(self):
        for json_file_type, path in self:
            print(f"Resolving: {json_file_type.__name__!r} {path!r}")
            namespace_file: NamespaceProxy = self.ctx.data[json_file_type]  # type: ignore
            smithed_file = SmithedJsonFile.parse_obj(namespace_file[path].data)

            if entries := smithed_file.smithed.entries():
                entries[-1].id = self.ctx.data.mcmeta.data.get("id", "")

                self.handle_file(smithed_file, path, json_file_type)

    def handle_file(
        self, file: SmithedJsonFile, path: str, json_file_type: type[JsonFile]
    ):
        models = {model.id: model for model in file.smithed.entries()}
        self.convert_to_after(models)
        resolved, processing = set(), []
        for current in models.keys():
            file = self.resolve(file, current, models, resolved, processing)
        self.ctx.data[json_file_type][path] = json_file_type(
            json.loads(file.json(by_alias=True, exclude={"smithed"}))
        )

    def convert_to_after(self, models: dict[str, SmithedModel]):
        for current_id, model in models.items():
            if before := model.priority.before:
                for id in before:
                    models[id].priority.after.add(current_id)
                model.priority.before.clear()

    def resolve(
        self,
        state: SmithedJsonFile,
        current: str,
        models: dict[str, SmithedModel],
        resolved: set[str],
        processing: list[str],  # we need to maintain order
    ):
        if after := models[current].priority.after:
            for id in after:
                if id in processing:
                    raise DependencyError(
                        f"Dependency loop found while resolving `{current}`.\n"
                        "The following loop was discovered: "
                        + " -> ".join(f"`{id}`" for id in processing)
                        + f" -> `{id}`"
                    )
                elif id in models:
                    processing.append(id)
                    state = SmithedJsonFile.parse_obj(
                        self.resolve(state, id, models, resolved, processing)
                    )
                    processing.remove(id)

        if current not in resolved:
            print(f"  Applying rules from pack: {current}")
            resolved.add(current)
            return SmithedJsonFile.parse_obj(self.apply_rule(state, models[current]))

        return state

    def apply_rule(self, state: SmithedJsonFile, model: SmithedModel):
        raw = state.dict(by_alias=True)
        for rule in model.rules:
            parsed_target = parse(rule.target)
            try:
                root = parsed_target.find(raw).pop(0)
            except IndexError:
                logger.warn(f"Target Path: {rule.target} was not found. Ignoring Rule.")
                continue

            match rule:
                case SmithedMergeRule(source=SmithedValueSource(value=value)):
                    match root.value:
                        case {} if type(value) is dict:
                            root.value |= value
                        case [] if type(value) is list:
                            root.value.extend(value)
                        case _ if type(value) not in {dict, list}:
                            parsed_target.update(raw, value)
                        case _:
                            raise WeldMergeError(
                                f"Unable to merge weld rules from pack {model.id}\n"
                                f"Current `{type(value)}` is not `{type(root.value)}`\n"
                            )

                case SmithedAppendRule(source=SmithedValueSource(value=value)):
                    root.value.append(value)

                case SmithedPrependRule(source=SmithedValueSource(value=value)):
                    root.value.prepend(value)

                case SmithedInsertRule(
                    source=SmithedValueSource(value=value), index=index
                ):
                    root.value.insert(index, value)

                case SmithedRemoveRule():
                    root.value.remove

        return raw

    def __iter__(self):
        for json_file_type, paths in self.cache.items():
            yield from [(json_file_type, path) for path in paths]


def setup(ctx: Context):
    handler = ctx.inject(ConflictsHandler)
    ctx.data.merge_policy.extend_namespace(LootTable, handler)


def beet_default(ctx: Context):
    handler = ctx.inject(ConflictsHandler)
    handler.handle_files()



from tokenstream import Token, TokenStream
from typing import Any

JsonDict = dict[str, Any]



def parse(raw: str):
    stream = TokenStream(raw)
    with stream.syntax(key=r"\w+", separator=r"\.", open=r"\[", close=r"\]", index=r"\[\d+\]"):
        for token in stream.collect():
            match token:
                case Token(type="open"):
                    yield stream.expect("index")
                    stream.expect("close")
                case Token(type=type) if type != "separator":
                    yield token


def get(path: str, obj: JsonDict):
    current = obj
    for token in parse(path):
        match token:
            case Token(type=type, value=str(value)) if type != "separator":
                current = current[value]
    return current

def set(path: str, obj: JsonDict, value: Any):
    current = obj
    for token in parse(path):
        match token:
            case Token(type=type, value=str(value)) if type != "separator":
                if value not in current:
                    current = [None] * (int(value) + 1) if 
                current = current[value]
    return current



import logging
from typing import Annotated, Any, Iterable, Literal

from beet import ListOption
from jsonpath_ng import parse
from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field, root_validator, validator

logger = logging.getLogger(__name__)


class BaseModel(PydanticBaseModel):
    class Config:
        json_encoders = {
            set: lambda v: list(v),
        }


class ConditionPackCheck(BaseModel):
    type: Literal["smithed:pack_check"]
    id: str


class ConditionInverted(BaseModel):
    type: Literal["smithed:inverted"]
    condition: "SmithedCondition"


SmithedCondition = ConditionPackCheck | ConditionInverted


class SmithedReferenceSource(BaseModel):
    type: Literal["smithed:reference"]
    path: str


class SmithedValueSource(BaseModel):
    type: Literal["smithed:value"] = "smithed:value"
    value: Any


SmithedSource = SmithedReferenceSource | SmithedValueSource


class SmithedBaseRule(BaseModel):
    target: str
    conditions: list[SmithedCondition] = []

class SmithedAdditiveRule(SmithedBaseRule):
    source: SmithedSource = Field(..., discriminator="type")

class SmithedMergeRule(SmithedAdditiveRule):
    type: Literal["smithed:merge"]

class SmithedAppendRule(SmithedAdditiveRule):
    type: Literal["smithed:append"]

class SmithedPrependRule(SmithedAdditiveRule):
    type: Literal["smithed:prepend"]

class SmithedInsertRule(SmithedAdditiveRule):
    type: Literal["smithed:insert"]
    index: int

class SmithedReplaceRule(SmithedAdditiveRule):
    type: Literal["smithed:replace"]

class SmithedRemoveRule(SmithedBaseRule):
    type: Literal["smithed:remove"]


SmithedRule = Annotated[
    SmithedMergeRule | SmithedAppendRule | SmithedPrependRule | SmithedInsertRule,
    Field(..., discriminator="type"),
]


class SmithedPriority(BaseModel):
    after: set[str] = set()
    before: set[str] = set()

    @validator("*")
    def convert_fields(cls, value: Iterable[str]):
        return set(value)


class SmithedModel(BaseModel, extra="forbid"):
    id: str = ""
    rules: list[SmithedRule]
    priority: SmithedPriority = SmithedPriority()


class SmithedJsonFile(BaseModel, extra="allow"):
    smithed: ListOption[SmithedModel] = Field(
        default_factory=ListOption, alias="__smithed__"
    )

    @root_validator
    def convert_type(cls, values: dict[str, ListOption[SmithedModel]]):
        for model in values["smithed"].entries():
            model.rules = list(cls.convert_rules(model.rules, values))

        return values
    
    @classmethod
    def convert_rules(cls, rules: list[SmithedRule], values: dict[str, Any]):
        for rule in rules:
            match rule.source:
                case SmithedReferenceSource(path=path):
                    try:
                        rule.source = SmithedValueSource(
                            value=parse(path).find(values).pop(0).value
                        )
                    except IndexError:
                        logger.warn(f"Source Reference Path: {path} was not found. Deleting Rule.")
                        continue
            yield rule



import json
import time
from argparse import ArgumentParser
from ensurepip import bootstrap
from glob import glob
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any

import click
import mecha
from beet import Context, Project, ProjectConfig, run_beet
from beet.core.utils import FileSystemPath
from beet.toolchain.cli import beet, message_fence
from beet.toolchain.commands import build

pass_project = click.make_pass_decorator(Project)  # type: ignore


@beet.command()
@pass_project
@click.pass_context
@click.argument("packs", nargs=-1)
@click.option("-l", "--link", metavar="WORLD", help="Link the project to a world")
def weld(
    ctx: click.Context,
    project: Project,
    packs: tuple[str],
    link: str | None,
):
    """Weld data and resource packs together with all data and assets in one zip.

    See https://wiki.smithed.dev/weld more info!
    """

    packs = tuple(expand_globs(packs))

    if len(packs) < 2:
        click.echo(click.style("Need at least one pack to weld", fg="red"))
        return -1

    with message_fence(f"Welding packs:\n - " + "\n - ".join(packs)):
        config = bootstrap_config(packs)

        if Path("beet.yaml").exists():
            config["extend"] = "beet.yaml"  # TODO: JANK

        with NamedTemporaryFile("w+", suffix=".json") as fp:
            json.dump(config, fp, indent=2)
            fp.read()  # still avoids bugs
            project.config_path = fp.name
            return ctx.invoke(build, link=link)


def expand_globs(packs: tuple[str]):
    for pack in packs:
        yield from glob(pack)


# weld_command = beet.command()(weld)
# weld_alias = click.command()(weld)



from contextlib import contextmanager
from functools import partial
from importlib import resources
from typing import Iterable, Literal
from zipfile import ZipFile

from beet import Context, JsonFile, ProjectConfig, run_beet, subproject
from beet.core.utils import FileSystemPath, JsonDict
from jinja2 import Template

DESCRIPTION = "Merged by Smithed Weld"
FABRIC_MOD_TEMPLATE = Template(
    (resources.files("weld") / "fabric.mod.json.j2").read_text()
)


@contextmanager
def run_weld(
    packs: Iterable[str] | Iterable[ZipFile],
    config: FileSystemPath | ProjectConfig | JsonDict = {},
    directory: FileSystemPath | None = None,
    pack_types: list[Literal["data_pack", "resource_pack"]] = ["data_pack"],
):
    if type(config) is dict:
        config |= {
            "output": "dist",
        }
        for pack_type in pack_types:
            config[pack_type] = {
                "zipped": True,
                "description": DESCRIPTION,
                "name": "welded-pack",
            }

    with run_beet(config, directory=directory) as ctx:
        ctx.require(partial(load_packs, packs=list(packs), pack_types=pack_types))

        yield ctx


def load_packs(
    ctx: Context,
    packs: list[str] | list[ZipFile],
    pack_types: list[Literal["data_pack", "resource_pack"]],
):
    # ctx.require("weld.setup")
    for pack in packs:
        for pack_type in pack_types:
            match pack:
                case str(name):
                    ctx.require(
                        subproject(
                            {
                                pack_type: {"load": name},
                                "pipeline": ["weld.print_pack_name"],
                            }
                        )
                    )
                case ZipFile() as file:
                    if pack_type == "data_pack":
                        ctx.data.load(file)
                    else:
                        ctx.assets.load(file)

    if len(pack_types) > 1:
        JsonFile(FABRIC_MOD_TEMPLATE.render(
            pack_hash=hash(tuple(packs)),
            pack_names=[
                pack.filename if type(pack) is ZipFile else pack for pack in packs
            ],
            mc_version=ctx.minecraft_version,
        ))
        ctx.require(partial(add_fabric_mod_json, packs=packs))


def add_fabric_mod_json(ctx: Context, packs: list[str] | list[ZipFile]):
    ctx.data.extra["fabric.mod.json"] = JsonFile(
        FABRIC_MOD_TEMPLATE.render(
            packs=packs,
            pack_names=[
                pack.filename if type(pack) is ZipFile else pack for pack in packs
            ],
            mc_version=ctx.minecraft_version,
        )
    )


def print_pack_name(ctx: Context):
    """Prints the merging pack name"""

    if not ctx.data and not ctx.assets:
        return

    print(
        f"Merging {'data' if ctx.data else 'resource'} pack:"
        f" {ctx.data.name or ctx.assets.name or 'Unknown'}"
    )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'smithed'}

packages = \
['weld']

package_data = \
{'': ['*']}

install_requires = \
['beet>=0.82.1,<0.83.0',
 'click>=8.1.3,<9.0.0',
 'jsonpath-ng>=1.5.3,<2.0.0',
 'mecha>=0.59.2,<0.60.0']

entry_points = \
{'beet': ['commands = weld.commands'],
 'console_scripts': ['weld = weld.commands:weld']}

setup_kwargs = {
    'name': 'smithed',
    'version': '0.7.1',
    'description': "Smithed's Python client with CLI, weld and more",
    'long_description': '# smithed-python\n> The python package for Smithed. Including weld, cli, and libraries (via [`smithed-libraries`](https://github.com/Smithed-MC/Libraries)).\n> **HEAVY WIP**\n',
    'author': 'Smithed Team',
    'author_email': 'team@smithed.dev',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import time
from io import BytesIO
from pathlib import Path
from zipfile import ZipFile

import streamlit as st
from streamlit.delta_generator import DeltaGenerator
from weld import run_weld

st.set_page_config(page_title="Smithed • Weld", page_icon="random")

st.write(
    '<h1 style="text-align: center; color: #104db4;">Weld (beta)</h1>',
    unsafe_allow_html=True,
)


st.write(
    "Merge all your data and resource packs!"
    " This tool will intelligently merge your packs to produce an easy to use zip."
    " It will also handle the following conflict types:\n"
    " - Tag files will merge exactly the same as vanilla\n"
    " - Lang files will extend each other\n"
    " - Model files will extend and sort overrides with `custom_model_data` defined\n"
    " - (Future) Custom `__smithed__` definitions will apply (see [vanilla-overides](https://wiki.smithed.dev/conventions/vanilla-overrides))\n"
)
st.warning(
    "⚠️ This tool is in **heavy beta** and should be used with caution.\n\nPlease forward all feedback to our [Discord](https://discord.com)!"
)


def upload_flow(label, pack_type: str, tab: DeltaGenerator):
    progress = tab.container()
    if pack_type == "all":
        tab.warning("🧪 This feature is experimental, use with caution!")
    packs: list[BytesIO] = tab.file_uploader(f"Upload {label}", accept_multiple_files=True, type="zip")  # type: ignore
    col1, col2 = tab.columns(2)
    t0 = time.perf_counter()
    path = None
    if col1.button("Build Packs", disabled=not packs, key=f"build-{label}"):
        progress.info(f"Building {len(packs)} packs!")
        pack_types = (
            ["data_pack", "resource_pack"] if pack_type == "all" else [pack_type]
        )
        with run_weld((ZipFile(pack) for pack in packs), pack_types=pack_types) as ctx:  # type: ignore
            match pack_type:
                case "data_pack":
                    path = ctx.data.save(path="output", zipped=True, overwrite=True)
                case "resource_pack":
                    path = ctx.assets.save(path="output", zipped=True, overwrite=True)
                case "all":
                    with ZipFile("output.jar", "w") as jar:
                        ctx.data.dump(jar)
                        ctx.assets.dump(jar)
                        path = Path(str(jar.filename))

        t1 = time.perf_counter()
        progress.success(f"Merged in {t1 - t0: 0.2f}s")
        if path is not None:
            with path.open(mode="rb") as file:
                col2.download_button(
                    "Download Welded Packs",
                    file,
                    file_name="welded-pack.zip"
                    if pack_type != "all"
                    else "welded-mod.jar",
                )
    else:
        col2.button("Download Welded Packs", disabled=True, key=f"download-{label}")

    if not packs:
        progress.info("Waiting for uploaded packs")
    elif path is None:
        progress.success("Ready to Build!")


data_col, resource_col, combined_col = st.tabs(
    ["Data Packs", "Resource Packs", "Combined (Fabric Jar)"]
)
upload_flow("Data Packs", "data_pack", data_col)
upload_flow("Resource Packs", "resource_pack", resource_col)
upload_flow("Any Packs", "all", combined_col)

st.write("---")

col1, col2 = st.columns(2, gap="medium")
col1.markdown(
    """
    <h4 style="text-align: center; color: #104db4;">Built with ♥️ by Smithed</h4>
    Smithed is a open-source ecosystem to support data and resource pack creators. Check us out at <a href="https://smithed.dev">here</a>!
    <br><br>
    Check out our <a href="https://wiki.smithed.dev/weld">documentation</a> to learn more about weld!
    """,
    unsafe_allow_html=True,
)

col2.markdown(
    """
    <iframe
     src="https://discord.com/widget?id=511303648119226382&amp;theme=dark"
     title="Discord Embed"
     sandbox="allow-popups allow-popups-to-escape-sandbox allow-same-origin allow-scripts"
     style="width: 325px; height: 400px; border: 2px solid white; border-radius: 16px;">
    </iframe>
    """,
    unsafe_allow_html=True,
)
